from distutils.util import strtobool
from datetime import datetime
from types import FunctionType
from dataclasses import fields
from typing import Dict
# Key points:
# 0) For get_instance_from_dict or get_dict_from_instance to work, the obj MUST be a dataclass!
# 1) conversion_fxn_map is a map from the STRING name of a data_type (e.g. 'int', not int)
#       to a function that converts an input value to that data_type.
# 2) The input values for get_instance_from_dict are currently all strings because they are
#       just payloads from http responses, but you can support any input data_type with this
#       framework. You simply have to define the conversion fxns to support the desired
#       input/output types.
# 3) Currently, get_dict_from_instance does not perform any type conversions. I may choose
#       to add that capability in the future, but it currently only converts the object to a
#       dict in the form {field_name_1 : field_val_1, ...}
class DataConverter:
    conversion_fxn_map : dict #[str : FunctionType]

    def __init__(self,extended_conversion_fxn_map=None):
        self.conversion_fxn_map = {cls.__name__: cls for cls in [str, int, float, bool, dict]}
        self.conversion_fxn_map['bool']=strtobool #allows us to convert the string input to a bool in a more intuitive way
        self.conversion_fxn_map['datetime']=lambda x : datetime.fromisoformat(x)

        if extended_conversion_fxn_map is not None:
            self.conversion_fxn_map.update(extended_conversion_fxn_map)
    # convert a single value to the specified data_type_str by using the appropriate conversion_fxn
    # e.g. convert_from('123','int') returns the integer-type 123. Note: 'int', not int.
    def convert_from(self,val,date_type_str):
        return self.conversion_fxn_map[date_type_str](val)

    # Converts a dict in the form {field_name_1 : field_val_1, ...} to an instance of a dataclass.
    # The dict may contain "extra" field_names not defined in the dataclass. Only the fields that
    # exist in the dataclass will be copied into the new instance object.
    def get_instance_from_dict(self,some_dict : dict, cls):
        converted_dict = {}
        for field in fields(cls):
            val, type_name=None, None
            if field.name in some_dict:
                val = some_dict[field.name]

                type_name=field.type

                # for some odd reason, type_name is sometimes a string literal and sometimes a class object
                # (e.g. sometimes the string 'int' and sometimes the class int)
                if not isinstance(type_name,str):
                    type_name = type_name.__name__

            if val is not None:
                converted_dict[field.name]=self.conversion_fxn_map[type_name](val)
            else:
                converted_dict[field.name]=None
        return cls(**converted_dict)

    def get_instances_from_arr_of_dicts(self,arr_of_dicts,cls):
        return [self.get_instance_from_dict(d,cls) for d in arr_of_dicts]
    # Convert a dataclass instance to a dict in the form {field_name_1 : field_val_1, ...}.
    # No type conversions are currently performed, but they can be added as needed.
    # Specify an "only_include" array to specify a subset of field_names to copy into the new dict
    def get_dict_from_instance(self,obj,only_include=None):
        if only_include is None:
            only_include = [f.name for f in fields(obj)]
        return {field_name : getattr(obj,field_name) for field_name in only_include}


from __future__ import annotations
from datetime import datetime
from dataclasses import dataclass,asdict

from jbg_util import binance_models #SocketData,OrderSocketData,PriceSocketData,BalanceSocketData,AssetBalance
from jbg_util import binance_util

binance_timestamp_to_dte,dte_to_binance_timestamp = binance_util.binance_timestamp_to_dte,binance_util.dte_to_binance_timestamp
SocketData,OrderSocketData,PriceSocketData,BalanceSocketData,AssetBalance = \
    binance_models.SocketData,binance_models.OrderSocketData,binance_models.PriceSocketData,binance_models.BalanceSocketData,binance_models.AssetBalance

class AssetSymbol:
    BTC = 'BTC'
    USD = 'USD'

def convert_socket_to_trader_data(data):
    if isinstance(data,BalanceSocketData):
        return BalanceTraderData.from_socket_data(data)
    elif isinstance(data,PriceSocketData):
        return PriceTraderData.from_socket_data(data)
    elif isinstance(data,OrderSocketData):
        return OrderTraderData.from_socket_data(data)
    raise ValueError("unrecognized datatype: ",type(data))

class OrderSide:
    buy = 'BUY'
    sell = 'SELL'

@dataclass
class OrderStatus:
    filled = 'FILLED'
    new = 'NEW'
    partially_filled = 'PARTIALLY_FILLED'
    canceled = 'CANCELED'
    pending_cancel = 'PENDING_CANCEL'
    rejected = 'REJECTED'
    expired = 'EXPIRED'

class OrderType:
    limit = 'LIMIT'
    market = 'MARKET'

@dataclass
class TraderData:
    event_type : str
    event_datetime : datetime

    @staticmethod
    def from_socket_data(socket_data : SocketData) -> TraderData:
        return TraderData(
            event_type=socket_data.e,
            event_datetime=binance_timestamp_to_dte(socket_data.E)
        )

@dataclass
class PriceTraderData(TraderData):
    symbol : str
    price : float

    def to_socket_data(self) -> PriceSocketData:
        return PriceSocketData(
            E=dte_to_binance_timestamp(self.event_datetime),
            e=self.event_type,
            s=self.symbol,
            c=self.price
        )

    @staticmethod
    def from_socket_data(socket_data : PriceSocketData) -> PriceTraderData:
        trader_data = TraderData.from_socket_data(socket_data)
        return PriceTraderData(
            **asdict(trader_data),
            symbol=socket_data.s,
            price = socket_data.c
        )

@dataclass
class OrderTraderData(TraderData):
    symbol : str
    order_id : int
    order_type : str
    side : str
    qty : float
    filled_qty : float
    price : float
    status : str
    create_datetime : datetime

    def to_socket_data(self) -> OrderSocketData:
        return OrderSocketData(
            E=dte_to_binance_timestamp(self.event_datetime),
            e=self.event_type,
            s=self.symbol,
            i=self.order_id,
            S=self.side,
            q=self.qty,
            z=self.filled_qty,
            p=self.price,
            X=self.status,
            o=self.order_type,
            O=dte_to_binance_timestamp(self.create_datetime)
        )

    @staticmethod
    def from_socket_data(socket_data : OrderSocketData) -> OrderTraderData:
        trader_data = TraderData.from_socket_data(socket_data)
        return OrderTraderData(
            **asdict(trader_data),
            symbol=socket_data.s,
            order_id=socket_data.i,
            side=socket_data.S,
            qty=socket_data.q,
            filled_qty=socket_data.z,
            price=socket_data.p,
            status=socket_data.X,
            order_type=socket_data.o,
            create_datetime=binance_timestamp_to_dte(socket_data.O)
        )

@dataclass
class TraderAssetBalance:
    free : float
    locked : float

@dataclass
class BalanceTraderData(TraderData):
    btc_free : float
    btc_locked : float
    usd_free : float
    usd_locked : float

    def to_socket_data(self) -> BalanceSocketData:
        return BalanceSocketData(
            E=dte_to_binance_timestamp(self.event_datetime),
            e=self.event_type,
            B=[
                AssetBalance(
                    a = AssetSymbol.BTC,
                    f= self.btc_free,
                    l=self.btc_locked
                ),
                AssetBalance(
                    a = AssetSymbol.USD,
                    f=self.usd_free,
                    l=self.usd_locked
                )
            ]
        )

    @staticmethod
    def from_socket_data(socket_data : BalanceSocketData) -> BalanceTraderData:
        btc_qtys = None
        usd_qtys = None
        for bal in socket_data.B:
            if bal.a=='BTC':
                btc_qtys=[bal.f, bal.l]
            elif bal.a=='USD':
                usd_qtys=[bal.f, bal.l]

        trader_data = TraderData.from_socket_data(socket_data)
        return BalanceTraderData(
            **asdict(trader_data),
            btc_free=btc_qtys[0],
            btc_locked=btc_qtys[1],
            usd_free=usd_qtys[0],
            usd_locked=usd_qtys[1]
        )








from __future__ import annotations
from typing import Generic, TypeVar
from dataclasses import dataclass

T = TypeVar("T")

@dataclass
class CircularBufferNode(Generic[T]):
    data : T
    next : CircularBufferNode = None

    def __repr__(self):
        return repr(self.data)

@dataclass
class CircularBuffer(Generic[T]):
    capacity : int
    curr_size : int = 0
    oldest : CircularBufferNode[T] = None
    latest : CircularBufferNode[T] = None

    def add(self,new_data : T):
        new_node = CircularBufferNode(new_data)
        if self.latest is not None:
            self.latest.next = new_node
        self.latest = new_node
        self.curr_size+=1
        change_in_size=self._update_oldest()
        self.curr_size+=change_in_size

    def _update_oldest(self):
        if self.oldest is None:
            self.oldest=self.latest
        elif self.curr_size>self.capacity:
            self.oldest = self.oldest.next
            return -1
        return 0

    def __iter__(self):
        curr_node = self.oldest
        while curr_node is not None:
            yield curr_node
            curr_node = curr_node.next

    def __str__(self):
        return ', '.join([str(node) for node in self])









from dataclasses import fields
from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String,DateTime,Float
from jbg_util import data_converter
from jbg_util import trader_models
from jbg_util import binance_models
from jbg_util import logging_util
from jbg_util import pub_sub_mediator

SocketChannel = binance_models.SocketChannel
Logger = logging_util.Logger
PubSubMediator = pub_sub_mediator.PubSubMediator
convert_socket_to_trader_data = trader_models.convert_socket_to_trader_data

class Database:
    TYPE_MAP = {
        'int' : Integer,
        'str' : String,
        'datetime' : DateTime,
        'float' : Float
    }

    def __init__(self,mediator : PubSubMediator = None,engine=None,logger : Logger = None):
        self.subscribe_for_updates(mediator)
        self.logger = logger or Logger(file_path='./logs/database.log',should_print_to_screen=True)
        self.engine = engine or Database.create_engine()
        self.tables : dict[str:Table]= {}
        self.data_converter=data_converter.DataConverter()

    def subscribe_for_updates(self,mediator : PubSubMediator):
        if mediator is None:
            return
        mediator.subscribe_local_fxn(self.on_update,SocketChannel.price)
        mediator.subscribe_local_fxn(self.on_update,SocketChannel.order)
        mediator.subscribe_local_fxn(self.on_update,SocketChannel.balance)

    async def on_update(self,data,channel):
        self.logger.secondary(f'\nDatabase received published data from socket channel {channel}!')
        parsed_data = convert_socket_to_trader_data(data)
        self.logger.info(f'\tParsed_data: {parsed_data}')
        self.save_model(parsed_data)

    def connect(self):
        self.conn = self.engine.connect()
        return self

    def create_table_from_cls(self,cls):
        meta = MetaData()
        table_name = cls.__name__
        try:
            cols = [Column(field.name, Database.TYPE_MAP[field.type.__name__]) for field in fields(cls) if field.name!='id']
        except AttributeError: #TODO: not sure what's happening here, and no time to figure it out right now
            cols = [Column(field.name, Database.TYPE_MAP[field.type]) for field in fields(cls) if field.name!='id']

        table = Table(
            table_name, meta,
            Column('id', Integer, primary_key = True),
            *cols,
        )
        meta.create_all(self.engine)
        self.tables[table_name]=table

    def get_table_for_cls(self,cls) -> Table:
        table_name = cls.__name__
        if table_name not in self.tables:
            self.create_table_from_cls(cls)
        return self.tables[table_name]

    #TODO make this async
    def save_model(self,model):
        table = self.get_table_for_cls(type(model))
        d = self.data_converter.get_dict_from_instance(model)
        ins = table.insert().values(**d)
        rslt = self.conn.execute(ins)

    def row_to_model(self,row : dict,cls):
        if not hasattr(cls,'id'):
            row = {key:val for key,val in row.items() if key!='id'}
        return cls(**row)

    @staticmethod
    def create_engine(db_name='db'):
        engine = create_engine(f'sqlite:///{db_name}.db', echo = True)
        meta = MetaData()
        meta.create_all(engine)
        return engine

    def clear_table_for_cls(self,cls,db_name='db'):
        table=self.get_table_for_cls(cls)
        stmt = table.delete() #.where(t.c.id > id)
        self.conn.execute(stmt)

    def get_all_saved_models(self,cls):
        table = self.get_table_for_cls(cls)
        s = table.select() #.where(table.c.id>2)
        rslt = self.conn.execute(s)
        return [self.row_to_model(row,cls) for row in rslt]


from __future__ import annotations
from dataclasses import dataclass
from jbg_util import trader_models
from jbg_util import websocket_util
from jbg_util import pub_sub_mediator
from jbg_util import logging_util
from jbg_util import circular_buffer
#from jbg_util.trader_engine_states import RallyTraderEngineState,PostRallyTraderEngineState,DumpTraderEngineState,PostDumpTraderEngineState,WarmingUpTraderEngineState,StableTraderEngineState,TraderEngineStateName,TraderEngineState

CircularBuffer = circular_buffer.CircularBuffer
Logger = logging_util.Logger
PubSubMediator = pub_sub_mediator.PubSubMediator
SocketChannel,PriceSocketData,OrderSocketData,BalanceSocketData = websocket_util.SocketChannel,websocket_util.PriceSocketData,websocket_util.OrderSocketData,websocket_util.BalanceSocketData
PriceTraderData,OrderTraderData,BalanceTraderData,TraderAssetBalance,AssetSymbol = \
    trader_models.PriceTraderData,trader_models.OrderTraderData,trader_models.BalanceTraderData,trader_models.TraderAssetBalance,trader_models.AssetSymbol
class TraderEngineDataType:
    price = 1
    balance = 2
    order = 3
@dataclass
class TraderEngineStateName:
    warming_up='warming_up'
    rally='rally'
    post_rally='post_rally'
    dump='dump'
    post_dump = 'post_dump'
    stable='stable'

class TraderEngineState:
    state_name : str = None

    def __init__(self,data : TraderEngineData,logger : Logger):
        self.logger : Logger = logger
        self.data : TraderEngineData = data

    def start(self):
        self.logger.warning(f'Starting new trader state: {self.state_name}')

    def end(self):
        self.logger.warning(f'Ending trader state: {self.state_name}')

    def do(self):
        self.logger.info(f'Doing trader state: {self.state_name}. Latest price={self.latest_price}')

    def next_state(self) -> str:
        return self.state_name
        #raise NotImplementedError("Cannot execute next_state on a generic TraderEngineState.")

    @property
    def oldest_price(self):
        return self.data.prices.oldest.data.price

    @property
    def latest_price(self):
        return self.data.prices.latest.data.price

    def is_dump(self):
        if self.oldest_price-self.latest_price > self.data.prices.capacity:
            return True
        return False

    def is_rally(self):
        if self.latest_price-self.oldest_price > self.data.prices.capacity:
            return True
        return False

class RallyTraderEngineState(TraderEngineState):
    state_name : str = TraderEngineStateName.rally
class PostRallyTraderEngineState(TraderEngineState):
    state_name : str= TraderEngineStateName.post_rally
class DumpTraderEngineState(TraderEngineState):
    state_name : str= TraderEngineStateName.dump
class PostDumpTraderEngineState(TraderEngineState):
    state_name : str = TraderEngineStateName.post_dump
class WarmingUpTraderEngineState(TraderEngineState):
    state_name : str = TraderEngineStateName.warming_up

    def next_state(self) -> str:
        if self.data.prices.curr_size!=self.data.prices.capacity:
            return TraderEngineStateName.warming_up
        if self.is_dump():
            return TraderEngineStateName.dump
        if self.is_rally():
            return TraderEngineStateName.rally
        return TraderEngineStateName.stable

class StableTraderEngineState(TraderEngineState):
    state_name : str = TraderEngineStateName.stable

    def next_state(self) -> str:
        if self.is_dump():
            return TraderEngineStateName.dump
        if self.is_rally():
            return TraderEngineStateName.rally
        return TraderEngineStateName.stable

class TraderEngine:
    data : TraderEngineData
    curr_state : TraderEngineState
    states : dict #[str : TraderEngineState]
    logger : Logger

    def __init__(self, mediator : PubSubMediator,logger : Logger = None, max_num_prices : int= 10):
        self.subscribe_for_updates(mediator)
        self.logger = logger or Logger('./logs/trader_engine.log',should_print_to_screen=True,log_in_color=True).empty_log()
        self.data = TraderEngineData(self.logger,max_num_prices)

        self.states = {
            cls.state_name : cls(self.data,self.logger) for cls in [RallyTraderEngineState,PostRallyTraderEngineState,
                DumpTraderEngineState,PostDumpTraderEngineState,WarmingUpTraderEngineState,StableTraderEngineState]
        }
        self.curr_state=self.states[TraderEngineStateName.warming_up]

    # def get_trader_state_by_name(state_name : str):

    def subscribe_for_updates(self,mediator : PubSubMediator):
        mediator.subscribe_local_fxn(self.on_price_update,SocketChannel.price)
        mediator.subscribe_local_fxn(self.on_order_update,SocketChannel.order)
        mediator.subscribe_local_fxn(self.on_balance_update,SocketChannel.balance)

    @property
    def prices(self) -> CircularBuffer[PriceTraderData]:
        return self.data.prices

    @property
    def balances(self) -> dict[str:TraderAssetBalance]:
        return self.data.balances

    @property
    def orders(self) -> dict[str:OrderTraderData]:
        return self.data.orders

    def change_state(self,new_state_name : str):
        self.logger.warning(f'Changing state from {self.curr_state.state_name} to {new_state_name}')
        new_state = self.states[new_state_name]

        self.curr_state.end()
        new_state.start()
        #new_state.do()

        self.curr_state = new_state

    def on_any_update(self):
        next_state=self.curr_state.next_state()
        if next_state != self.curr_state.state_name:
            # self.logger.warning(f'CHANGING STATE! {next_state}, {self.curr_state.state_name}')
            self.change_state(next_state)
        self.curr_state.do()

    async def on_price_update(self,data : PriceSocketData,channel):
        trader_data = PriceTraderData.from_socket_data(data)
        self.logger.success(f'Received price update: {trader_data}')
        self.data.on_price_update(trader_data)
        #self.data.prices.add(trader_data)
        self.on_any_update()

    async def on_order_update(self,data : OrderSocketData,channel):
        trader_data = OrderTraderData.from_socket_data(data)
        self.logger.success(f'Received order update: {trader_data}')
        self.data.on_order_update(trader_data)
        # self.orders[trader_data.order_id] = trader_data
        self.on_any_update()

    async def on_balance_update(self,data : BalanceSocketData,channel):
        trader_data = BalanceTraderData.from_socket_data(data)
        self.logger.success(f'Received balance update: {trader_data}')
        self.data.on_balance_update(trader_data)
        # self.balances[AssetSymbol.BTC] = TraderAssetBalance(free=trader_data.btc_free,locked=trader_data.btc_locked)
        # self.balances[AssetSymbol.USD] = TraderAssetBalance(free=trader_data.usd_free,locked=trader_data.usd_locked)
        self.on_any_update()

    # def print_buffer_sizes(self):
    #     for name,buffer in [('prices',self.prices),('orders',self.orders),('balances',self.balances)]:
    #         self.logger.warning(f'CircBuffer {name} has size {buffer.curr_size}')

    # def print_buffer_latests(self):
    #     for name,buffer in [('prices',self.prices),('orders',self.orders),('balances',self.balances)]:
    #         self.logger.warning(f'CircBuffer {name} has latest {buffer.latest}')

class TraderEngineData:
    prices : CircularBuffer[PriceTraderData]
    orders : dict #[int : OrderTraderData]
    balances : dict #[str : TraderAssetBalance]

    num_updates_by_type : dict[str : int]

    def __init__(self,logger : Logger, max_num_prices : int):
        self.logger = logger
        self.prices = CircularBuffer(capacity=max_num_prices)
        self.orders = {}
        self.balances = {}
        self.num_updates_by_type = {t:0 for t in [TraderEngineDataType.price,TraderEngineDataType.order,TraderEngineDataType.balance]}

    @property
    def btc(self):
        return self.balances[AssetSymbol.BTC]

    @property
    def usd(self):
        return self.balances[AssetSymbol.USD]

    def on_price_update(self,data : PriceTraderData):
        self.prices.add(data)
        self.num_updates_by_type[TraderEngineDataType.price]+=1

    def on_order_update(self,data : OrderTraderData):
        self.orders[data.order_id] = data
        self.num_updates_by_type[TraderEngineDataType.order]+=1

    def on_balance_update(self,data : BalanceTraderData):
        self.balances[AssetSymbol.BTC] = TraderAssetBalance(free=data.btc_free,locked=data.btc_locked)
        self.balances[AssetSymbol.USD] = TraderAssetBalance(free=data.usd_free,locked=data.usd_locked)
        self.num_updates_by_type[TraderEngineDataType.balance]+=1

